@@ -1,7 +1,7 @@
 # +--------------------------------------------------------------------------+
 # |  Licensed Materials - Property of IBM                                    |
 # |                                                                          |
-# | (C) Copyright IBM Corporation 2008, 2014.                                |
+# | (C) Copyright IBM Corporation 2008, 2016.                                |
 # +--------------------------------------------------------------------------+
 # | This module complies with SQLAlchemy 0.8 and is                          |
 # | Licensed under the Apache License, Version 2.0 (the "License");          |
@@ -33,7 +33,7 @@ from sqlalchemy.types import BLOB, CHAR, CLOB, DATE, DATETIME, INTEGER,\
     SMALLINT, BIGINT, DECIMAL, NUMERIC, REAL, TIME, TIMESTAMP,\
     VARCHAR, FLOAT
 
-SA_Version = [long(ver_token) for ver_token in SA_Version.split('.')[0:2]]
+SA_Version = [int(ver_token) for ver_token in SA_Version.split('.')[0:2]]
 
 # as documented from:
 # http://publib.boulder.ibm.com/infocenter/db2luw/v9/index.jsp?topic=/com.ibm.db2.udb.doc/admin/r0001095.htm
@@ -301,7 +301,11 @@ class DB2TypeCompiler(compiler.GenericTypeCompiler):
         return self.visit_FLOAT(type_)
 
     def visit_unicode(self, type_):
-        return self.visit_VARGRAPHIC(type_)
+        check_server = getattr(DB2Dialect,'serverType')
+        if(check_server == "DB2"):
+            return (self.visit_VARGRAPHIC(type_) + " CCSID 1200")
+        else:
+            return self.visit_VARGRAPHIC(type_)
 
     def visit_unicode_text(self, type_):
         return self.visit_LONGVARGRAPHIC(type_)
@@ -325,6 +329,9 @@ class DB2Compiler(compiler.SQLCompiler):
         def visit_true(self, expr, **kw):
             return '1'
             
+    def get_cte_preamble(self, recursive):
+        return "WITH"
+        
     def visit_now_func(self, fn, **kw):
         return "CURRENT_TIMESTAMP"
 
@@ -335,12 +342,12 @@ class DB2Compiler(compiler.SQLCompiler):
             return ' WITH RS USE AND KEEP SHARE LOCKS'
         else:
             return ''
-
+			
     def visit_mod_binary(self, binary, operator, **kw):
         return "mod(%s, %s)" % (self.process(binary.left),
                                                 self.process(binary.right))
 
-    def limit_clause(self, select):
+    def limit_clause(self, select,**kwargs):
         if (select._limit is not None) and (select._offset is None):
             return " FETCH FIRST %s ROWS ONLY" % select._limit
         else:
@@ -433,13 +440,13 @@ class DB2Compiler(compiler.SQLCompiler):
         # for example
         if isinstance(type_, (
                     sa_types.DateTime, sa_types.Date, sa_types.Time,
-                    sa_types.DECIMAL)):
+                    sa_types.DECIMAL,sa_types.String)):
             return super(DB2Compiler, self).visit_cast(cast, **kw)
         else:
             return self.process(cast.clause)
 
-    def get_select_precolumns(self, select):
-        if isinstance(select._distinct, basestring):
+    def get_select_precolumns(self, select,**kwargs):
+        if isinstance(select._distinct, str):
             return select._distinct.upper() + " "
         elif select._distinct:
             return "DISTINCT "
@@ -474,10 +481,29 @@ class DB2Compiler(compiler.SQLCompiler):
             return super(DB2Compiler, self).visit_unary(unary, **kw)
 
 class DB2DDLCompiler(compiler.DDLCompiler):
+    
+    def get_server_version_info(self, dialect):
+        """Returns the DB2 server major and minor version as a list of ints."""
+        if hasattr(dialect, 'dbms_ver'):
+            return [int(ver_token) for ver_token in dialect.dbms_ver.split('.')[0:2]]
+        else:
+            return []
+    
+    def _is_nullable_unique_constraint_supported(self, dialect):
+        """Checks to see if the DB2 version is at least 10.5.
+        This is needed for checking if unique constraints with null columns are supported.
+        """
+
+        dbms_name = getattr(dialect, 'dbms_name', None)
+        if hasattr(dialect, 'dbms_name'):
+           if dbms_name != None and (dbms_name.find('DB2/') != -1):
+                return self.get_server_version_info(dialect) >= [10, 5]
+        else:
+            return False
 
     def get_column_specification(self, column, **kw):
         col_spec = [self.preparer.format_column(column)]
-        col_spec.append(self.dialect.type_compiler.process(column.type))
+        col_spec.append(self.dialect.type_compiler.process(column.type,type_expression=column))
 
 
         # column-options: "NOT NULL"
@@ -512,37 +538,48 @@ class DB2DDLCompiler(compiler.DDLCompiler):
     def visit_drop_constraint(self, drop, **kw):
         constraint = drop.element
         if isinstance(constraint, sa_schema.ForeignKeyConstraint):
-                qual = "FOREIGN KEY "
-                const = self.preparer.format_constraint(constraint)
+            qual = "FOREIGN KEY "
+            const = self.preparer.format_constraint(constraint)
         elif isinstance(constraint, sa_schema.PrimaryKeyConstraint):
-                qual = "PRIMARY KEY "
-                const = ""
+            qual = "PRIMARY KEY "
+            const = ""
         elif isinstance(constraint, sa_schema.UniqueConstraint):
-                qual = "INDEX "
-                const = self.preparer.format_constraint(constraint)
+            qual = "UNIQUE "
+            if self._is_nullable_unique_constraint_supported(self.dialect):
+                for column in constraint:
+                    if column.nullable:
+                        constraint.uConstraint_as_index = True
+                if getattr(constraint, 'uConstraint_as_index', None):
+                    qual = "INDEX "
+            const = self.preparer.format_constraint(constraint)
         else:
-                qual = ""
-                const = self.preparer.format_constraint(constraint)
+            qual = ""
+            const = self.preparer.format_constraint(constraint)
+            
+        if hasattr(constraint, 'uConstraint_as_index') and constraint.uConstraint_as_index:
+            return "DROP %s%s" % \
+                                (qual, const)
         return "ALTER TABLE %s DROP %s%s" % \
                                 (self.preparer.format_table(constraint.table),
                                 qual, const)
                                 
     def create_table_constraints(self, table, **kw):
-        if (self.dialect.dbms_name.find('DB2/') != -1) and ([long(ver_token) for ver_token in self.dialect.dbms_ver.split('.')[0:2]] >= [10, 5]):
+        if self._is_nullable_unique_constraint_supported(self.dialect):
             for constraint in table._sorted_constraints:
                 if isinstance(constraint, sa_schema.UniqueConstraint):
                     for column in constraint:
                         if column.nullable:
                             constraint.use_alter = True
+                            constraint.uConstraint_as_index = True
                             break
-                    if hasattr(constraint, 'use_alter') and constraint.use_alter:
+                    if getattr(constraint, 'uConstraint_as_index', None):
                         if not constraint.name:
                             index_name = "%s_%s_%s" % ('ukey', self.preparer.format_table(constraint.table), '_'.join(column.name for column in constraint))
                         else:
                             index_name = constraint.name
                         index = sa_schema.Index(index_name, *(column for column in constraint))
                         index.unique = True
-                        index.exclude_nulls = True
+                        index.uConstraint_as_index = True
         result = super( DB2DDLCompiler, self ).create_table_constraints(table, **kw)
         return result
     
@@ -551,25 +588,25 @@ class DB2DDLCompiler(compiler.DDLCompiler):
             sql = super( DB2DDLCompiler, self ).visit_create_index(create)
         else:
             sql = super( DB2DDLCompiler, self ).visit_create_index(create, include_schema, include_table_schema)
-        if getattr(create.element, 'exclude_nulls', None):
+        if getattr(create.element, 'uConstraint_as_index', None):
             sql += ' EXCLUDE NULL KEYS'
         return sql
 
     def visit_add_constraint(self, create):
-        if (self.dialect.dbms_name.find('DB2/') != -1) and ([long(ver_token) for ver_token in self.dialect.dbms_ver.split('.')[0:2]] >= [10, 5]):
+        if self._is_nullable_unique_constraint_supported(self.dialect):
             if isinstance(create.element, sa_schema.UniqueConstraint):
                 for column in create.element:
                     if column.nullable:
-                        create.element.exclude_nulls = True
+                        create.element.uConstraint_as_index = True
                         break
-                if getattr(create.element, 'exclude_nulls', None):
+                if getattr(create.element, 'uConstraint_as_index', None):
                     if not create.element.name:
                         index_name = "%s_%s_%s" % ('uk_index', self.preparer.format_table(create.element.table), '_'.join(column.name for column in create.element))
                     else:
                         index_name = create.element.name
                     index = sa_schema.Index(index_name, *(column for column in create.element))
                     index.unique = True
-                    index.exclude_nulls = True
+                    index.uConstraint_as_index = True
                     sql = self.visit_create_index(sa_schema.CreateIndex(index)) 
                     return sql
         sql = super( DB2DDLCompiler, self ).visit_add_constraint(create)
@@ -578,20 +615,12 @@ class DB2DDLCompiler(compiler.DDLCompiler):
 class DB2IdentifierPreparer(compiler.IdentifierPreparer):
 
     reserved_words = RESERVED_WORDS
-    illegal_initial_characters = set(xrange(0, 10)).union(["_", "$"])
-
-
-class DB2ExecutionContext(default.DefaultExecutionContext):
-    def fire_sequence(self, seq, type_):
-        return self._execute_scalar("SELECT NEXTVAL FOR " +
-                    self.dialect.identifier_preparer.format_sequence(seq) +
-                    " FROM SYSIBM.SYSDUMMY1", type_)
+    illegal_initial_characters = set(range(0, 10)).union(["_", "$"])
 
 class _SelectLastRowIDMixin(object):
     _select_lastrowid = False
     _lastrowid = None
 
-
     def get_lastrowid(self):
         return self._lastrowid
 
@@ -616,6 +645,13 @@ class _SelectLastRowIDMixin(object):
                 self._lastrowid = int(row[0])
 
 
+class DB2ExecutionContext(_SelectLastRowIDMixin,default.DefaultExecutionContext):
+    def fire_sequence(self, seq, type_):
+        return self._execute_scalar("SELECT NEXTVAL FOR " +
+                    self.dialect.identifier_preparer.format_sequence(seq) +
+                    " FROM SYSIBM.SYSDUMMY1", type_)
+
+
 class DB2Dialect(default.DefaultDialect):
 
     name = 'ibm_db_sa'
@@ -653,6 +689,7 @@ class DB2Dialect(default.DefaultDialect):
     execution_ctx_cls = DB2ExecutionContext
 
     _reflector_cls = ibm_reflection.DB2Reflector
+    serverType = ''
 
     def __init__(self, **kw):
         super(DB2Dialect, self).__init__(**kw)
@@ -662,9 +699,23 @@ class DB2Dialect(default.DefaultDialect):
     # reflection: these all defer to an BaseDB2Reflector
     # object which selects between DB2 and AS/400 schemas
     def initialize(self, connection):
+        self.dbms_ver = getattr(connection.connection, 'dbms_ver', None)
+        self.dbms_name = getattr(connection.connection, 'dbms_name', None)
+        DB2Dialect.serverType = self.dbms_name
         super(DB2Dialect, self).initialize(connection)
-        self.dbms_ver = connection.connection.dbms_ver
-        self.dbms_name = connection.connection.dbms_name
+		#check server type logic here
+        print(self.dbms_name)
+        if(self.dbms_name == 'AS'):
+            _reflector_cls = ibm_reflection.AS400Reflector
+        elif(self.dbms_name == "DB2"):
+            _reflector_cls = ibm_reflection.OS390Reflector
+        elif("DB2/" in self.dbms_name):
+            _reflector_cls = ibm_reflection.DB2Reflector
+        elif("IDS/" in self.dbms_name):
+            _reflector_cls = ibm_reflection.DB2Reflector
+        
+        self._reflector = _reflector_cls(self) 
+
         
     def normalize_name(self, name):
         return self._reflector.normalize_name(name)
