@@ -1,7 +1,7 @@
 # +--------------------------------------------------------------------------+
 # |  Licensed Materials - Property of IBM                                    |
 # |                                                                          |
-# | (C) Copyright IBM Corporation 2008, 2014.                                |
+# | (C) Copyright IBM Corporation 2008, 2016.                                |
 # +--------------------------------------------------------------------------+
 # | This module complies with SQLAlchemy 0.8 and is                          |
 # | Licensed under the Apache License, Version 2.0 (the "License");          |
@@ -13,14 +13,20 @@
 # | KIND, either express or implied. See the License for the specific        |
 # | language governing permissions and limitations under the License.        |
 # +--------------------------------------------------------------------------+
-# | Authors: Alex Pitigoi, Abhigyan Agrawal, Rahul Priyadarshi               |
-# | Contributors: Jaimy Azle, Mike Bayer                                     |
+# | Authors: Alex Pitigoi, Abhigyan Agrawal, Rahul Priyadarshi,Abhinav Radke |
+# | Contributors: Jaimy Azle, Mike Bayer,Hemlata Bhatt                       |
 # +--------------------------------------------------------------------------+
 
 from .base import DB2ExecutionContext, DB2Dialect
 from sqlalchemy import processors, types as sa_types, util
 from sqlalchemy import __version__ as SA_Version
-SA_Version = [long(ver_token) for ver_token in SA_Version.split('.')[0:2]]
+from sqlalchemy.exc import ArgumentError
+SA_Version = [int(ver_token) for ver_token in SA_Version.split('.')[0:2]]
+SQL_TXN_READ_UNCOMMITTED = 1
+SQL_TXN_READ_COMMITTED = 2
+SQL_TXN_REPEATABLE_READ = 4
+SQL_TXN_SERIALIZABLE = 8
+SQL_ATTR_TXN_ISOLATION = 108
 
 if SA_Version < [0, 8]:
     from sqlalchemy.engine import base
@@ -75,7 +81,7 @@ class DB2ExecutionContext_ibm_db(DB2ExecutionContext):
 class DB2Dialect_ibm_db(DB2Dialect):
 
     driver = 'ibm_db_sa'
-    supports_unicode_statements = False
+    supports_unicode_statements = True
     supports_sane_rowcount = True
     supports_sane_multi_rowcount = False
     supports_native_decimal = False
@@ -107,7 +113,48 @@ class DB2Dialect_ibm_db(DB2Dialect):
 
     def _get_server_version_info(self, connection):
         return connection.connection.server_info()
+    
+    _isolation_lookup = set(['READ STABILITY','RS', 'UNCOMMITTED READ','UR',
+                             'CURSOR STABILITY','CS', 'REPEATABLE READ','RR'])
+   
+    _isolation_levels_cli = {'RR': SQL_TXN_SERIALIZABLE, 'REPEATABLE READ': SQL_TXN_SERIALIZABLE, 
+                            'UR': SQL_TXN_READ_UNCOMMITTED, 'UNCOMMITTED READ': SQL_TXN_READ_UNCOMMITTED,
+                             'RS': SQL_TXN_REPEATABLE_READ, 'READ STABILITY': SQL_TXN_REPEATABLE_READ,   
+                             'CS': SQL_TXN_READ_COMMITTED, 'CURSOR STABILITY': SQL_TXN_READ_COMMITTED }
+    
+    _isolation_levels_returned = { value : key for key, value in _isolation_levels_cli.items()}
 
+    def _get_cli_isolation_levels(self, level):
+        return _isolation_levels_cli[level]
+
+    def set_isolation_level(self, connection, level):    
+        if level is  None:
+         level ='CS' 
+        else :
+          if len(level.strip()) < 1:
+            level ='CS'
+        level.upper().replace("-", " ")   
+        if level not in self._isolation_lookup:
+            raise ArgumentError(
+                "Invalid value '%s' for isolation_level. "
+                "Valid isolation levels for %s are %s" %
+                (level, self.name, ", ".join(self._isolation_lookup))
+            )
+        attrib = {SQL_ATTR_TXN_ISOLATION:_get_cli_isolation_levels(self,level)}
+        res = connection.set_option(attrib)
+
+        
+    def get_isolation_level(self, connection):
+                
+        attrib = SQL_ATTR_TXN_ISOLATION
+        res = connection.get_option(attrib)
+
+        val = self._isolation_levels_returned[res]
+        return val
+    
+    def reset_isolation_level(self, connection):
+        self.set_isolation_level(connection,'CS')
+        
     def create_connect_args(self, url):
         # DSN support through CLI configuration (../cfg/db2cli.ini),
         # while 2 connection attributes are mandatory: database alias
@@ -132,15 +179,17 @@ class DB2Dialect_ibm_db(DB2Dialect):
             if url.username:
                 dsn_param.append('UID=%s' % url.username)
             if url.password:
+                if ';' in url.password:
+                    url.password=(url.password).partition(";")[0]
                 dsn_param.append('PWD=%s' % url.password)
             
-            #check for SSL arguments
-            ssl_keys = ['Security', 'SSLClientKeystoredb', 'SSLClientKeystash']
+            #check for connection arguments
+            connection_keys = ['Security', 'SSLClientKeystoredb', 'SSLClientKeystash','SSLServerCertificate']
             query_keys = url.query.keys()
-            for key in ssl_keys:
+            for key in connection_keys:
                 for query_key in query_keys:
                     if query_key.lower() == key.lower():
-                        dsn_param.append('%(ssl_key)s=%(value)s' % {'ssl_key': key, 'value': url.query[query_key]})
+                        dsn_param.append('%(connection_key)s=%(value)s' % {'connection_key': key, 'value': url.query[query_key]})
                         del url.query[query_key]
                         break
                            
@@ -157,10 +206,12 @@ class DB2Dialect_ibm_db(DB2Dialect):
     def is_disconnect(self, ex, connection, cursor):
         if isinstance(ex, (self.dbapi.ProgrammingError,
                                              self.dbapi.OperationalError)):
-            return 'Connection is not active' in str(ex) or \
-                        'connection is no longer active' in str(ex) or \
-                        'Connection Resource cannot be found' in str(ex) or \
-                        'SQL30081N' in str(ex)
+            connection_errors = ('Connection is not active', 'connection is no longer active',
+                                    'Connection Resource cannot be found', 'SQL30081N'
+                                    'CLI0108E', 'CLI0106E', 'SQL1224N')
+            for err_msg in connection_errors:
+                if err_msg in str(ex):
+                    return True
         else:
             return False
 
